/*
 * config.c
 *
 *  Created on: Jul 27, 2024
 *      Author: true
 */

#include <string.h>
#include <ch32v20x.h>

#include "config.h"
#include "flash.h"



struct UserConf userconf;

static uint8_t active_page;



uint32_t chip_get_flash_size()
{
    uint32_t w;

    w   = DBGMCU_GetCHIPID();
    w >>= 16;
    w  &= 0xff;

    if (w >= 0x80) return 131072;   // can be 3 different sizes, but I don't care right now;
                                    // no applicable MCU for us has this size anyway
    switch (w) {
        case 0x33:
        case 0x36:
        case 0x37:
        case 0x38:
        case 0x39: {
            return 32768;
        }
        default: {
            return 65536;
        }
    }
}

static uint32_t calc_address()
{
    uint32_t calc = (FLASH_BASE + chip_get_flash_size()) - (CONF_FLASH_PAGE_SIZE * CONF_FLASH_PAGES);
    return calc;
}

/*
static void read_page_from_flash(uint8_t page, uint32_t *data, uint16_t len)
{
    uint32_t *addr = calc_address();

    // no more than 256 bytes at a time necessary
    if (len > 64) len = 64;

    // set address of page
    addr += page * CONF_FLASH_PAGE_SIZE;

    // read the data
    flash_read(addr, data, len);
}
*/

static void write_page_to_flash(uint8_t page, uint32_t *data, uint16_t len)
{
    len = (uint16_t)len;

    uint32_t addr = calc_address() + (page * CONF_FLASH_PAGE_SIZE);

    // write the data
    // note we don't pass any length. we'll just read whatever garbage
    // is in RAM after our config and write it to flash. lol
    flash_write256((uint32_t *)addr, data);
}

static uint16_t checksum(struct UserConf *conf)
{
    uint16_t i;
    uint16_t sum = 0;

    uint8_t *uc = (uint8_t *)conf;

    // calculate checksum
    for (i = 0; i < sizeof(userconf) - 6; i++) {
        sum += *uc++;
    }

    return sum;
}

void userconf_load()
{
    uint8_t i;

    uint32_t ver_highest = 0;

    uint8_t page = CONF_FLASH_PAGES;
    uint32_t addr;
    struct UserConf *flash;

    // read pages and see if we can find our data
    while (page--) {
        addr = calc_address() + (page * CONF_FLASH_PAGE_SIZE);
        flash = (struct UserConf *)addr;
        if (flash->checkval == CHECKVAL) {
            if (flash->checksum == checksum(flash)) {
                if (ver_highest < flash->version) {
                    ver_highest = flash->version;
                    active_page = page;
                }
            }
        }
    }

    if (!ver_highest) {
        // config is invalid; reset to default
        userconf.version = 0;

        // default program configs
        for (i = 0; i < 8; i++) {
            userconf.pep_conf[i][0] = 0;
            userconf.pep_conf[i][1] = 0;
            userconf.pep_conf[i][2] = 0;
            userconf.pep_conf[i][3] = 0;
        }

        // program 0: flashing
        userconf.pep_conf[0][0] = 0x80;     // brightness
        userconf.pep_conf[0][1] = 100;      // flash rate

        // program 2: loops
        userconf.pep_conf[2][0] = 0xc0;     // brightness
        userconf.pep_conf[2][1] = 1;        // looping rate
        userconf.pep_conf[2][3] = 2;        // decay rate

        // program 4: alternating
        userconf.pep_conf[4][0] = 0x80;     // brightness
        userconf.pep_conf[4][1] = 12;        // flash rate
        userconf.pep_conf[4][3] = 0x0b;     // lower brightness on alternate

        userconf.checksum = checksum(&userconf);
        userconf.checkval = CHECKVAL;
    } else {
        memcpy(&userconf, (uint8_t *)(calc_address() + (active_page * 256)), sizeof(userconf));
    }
}

void userconf_save()
{
    // we can only save ~4 billion times
    userconf.version++;
    userconf.checksum = checksum(&userconf);
    userconf.checkval = CHECKVAL;

    // determine page to write
    active_page++;
    active_page %= CONF_FLASH_PAGES;

    // this MCU writes and erases in full page sized blocks.
    // nothing is mentioned in the datasheet nor reference manual
    // about repeated writes. so to be safe, each page gets
    // written to once per erase.
    write_page_to_flash(active_page, (uint32_t *)&userconf, sizeof(userconf)/4);
}
